==> storm-sources-0.003.tar.gz: SToRM/reference_data.c <==
#include "reference_data.h"
#include <ctype.h>
#include <unistd.h>
#include <errno.h> 

#include "alignment.h" /* for INDEL_COUNT_LIMIT */

int IGNORE_LOWERCASE = 0;

/**
 * Load reference sequence
 */

inline int load_reference_sequence(FILE* ref_in, ReferenceDBType* db) {
  char line[MAX_LINE_LEN];
  int masked_region_start = -1;
  MaskedRegionType* crt_masked_region = db->masked_regions;

  db->name = NULL;

  /* 1) skip (but keep the name of) the fasta header... */
  if (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) == LINE_TYPE_SEQ_HEADER) {
    int name_len = strcspn(line + 1, "\r\n");
    if (name_len > 0) {
      SAFE_FAILURE__ALLOC(db->name, (name_len + 1), char);
      db->name[name_len] = '\0';
      strncpy(db->name, line + 1, name_len);
    } else {
      ERROR__(("Empty fasta header line in the reference sequence.\n"));
      return RETURN_INPUT_ERR;
    }
  } else {
    ERROR__(("Invalid fasta header line in the reference sequence.\n"));
    return RETURN_INPUT_ERR;
  }


  /* 2) keep the start of the nucleotide sequence to come back */
  int i_start = ftell(ref_in);
  int size = 0;

  /* 3) read nucleotide data to measure its size */
  if (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) != LINE_TYPE_SEQ_HEADER) {
    // this is the first line of the sequence
    int line_len = strlen(line);
    
    /* ignore the newline */
    while (line_len > 0 && isspace(line[line_len - 1])) {
      --line_len;
    }

    /* when the first full line of data is empty : error */
    if (line_len <= 0) {
      ERROR__(("Empty dataline in the reference sequence.\n"));
      return RETURN_INPUT_ERR;
    }

#ifndef NUCLEOTIDES
    /* without the first base */
    size = line_len - 1;    
    db->first_base = BASE_SYMBOL_TO_CODE(line[0]);
#else
    /* with the first base */
    size = line_len; 
#endif
    
    /* 3.1) full nucleotide data loop to measure the size */
    while (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) != LINE_TYPE_SEQ_HEADER) {
      line_len = strlen(line);
      /* ignore the newline */
      while (line_len > 0 && isspace(line[line_len - 1])) {
	--line_len;
      }
      size += line_len;
    }

    /* 3.2) Allocate the estimated size (>= true size) */
    SAFE_FAILURE__ALLOC(db->sequence, COMPRESSED_LEN_N_BYTES(size+2*(INDEL_COUNT_LIMIT)+1,N_BYTES), CODE_TYPE);
    memset(db->sequence,'\0',COMPRESSED_LEN_N_BYTES(size+2*(INDEL_COUNT_LIMIT)+1,N_BYTES));
    /* 3.3) Go back to first base and load the nucleotides sequence */
    fseek(ref_in, i_start, SEEK_SET);

    char c = (IGNORE_LOWERCASE ? fgetc(ref_in) : toupper(fgetc(ref_in)));
    if (!IS_VALID_BASE_SYMBOL(c)) {
      ERROR__(("Invalid first symbol in the reference sequence.\n"));
      return RETURN_INPUT_ERR;

    }
    CODE_TYPE c1 = BASE_SYMBOL_TO_CODE(c);
    int  i = 0;
#ifdef NUCLEOTIDES
    TO_NTH_CODE(db->sequence, i, c1);
    i++;
#endif
    while ((c = (IGNORE_LOWERCASE ? fgetc(ref_in) : toupper(fgetc(ref_in)))) != EOF) {
      if (isspace(c)) {
	continue;
      }

      if (c == SEQ_HEADER_MARKER) {
	/* started to read the next sequence; getting out */
	fseek(ref_in, -1, SEEK_CUR);
	break;
      }
      CODE_TYPE c2;
      if (!IS_VALID_BASE_SYMBOL(c)) {
	c2 = c1; /* masked regions, eg 'NNNNNNNNN' cannot be coded; instead, GNNNNNNNN is transformed in GGGGGGGGG */
	if (masked_region_start == -1) {
	  /* beginning of a new masked region */
	  masked_region_start = i;
	}
      } else {
	c2 = BASE_SYMBOL_TO_CODE(c);
	if (masked_region_start != -1) {
	  /* a masked region ends here */
	  if (crt_masked_region == NULL) {
	    /* this is the first one */
	    SAFE_FAILURE__ALLOC(db->masked_regions, 1, MaskedRegionType);
	    crt_masked_region = db->masked_regions;
	  } else {
	    SAFE_FAILURE__ALLOC(crt_masked_region->next, 1, MaskedRegionType);
	    crt_masked_region = crt_masked_region->next;
	  }
	  crt_masked_region->start = masked_region_start;
	  crt_masked_region->end = i;
	  masked_region_start = -1;
	}
      }
#ifdef NUCLEOTIDES
      TO_NTH_CODE(db->sequence, i, c2);
#else
      TO_NTH_CODE(db->sequence, i, COLOR(c1, c2));
      c1 = c2;
#endif
      i++;
    }
    db->size = i;
    return db->size;
  } else {
    ERROR__(("Empty fasta block in the reference sequence.\n"));
    return RETURN_INPUT_ERR;
  }
}

/**
 * Load reference sequence
 */
int load_reference_db(const char* reference_filename, ReferenceDBType** db) {
  char line[MAX_LINE_LEN];
  FILE* ref_in = fopen(reference_filename, "r");
  int i=0;
  int sequence_count = 0;
  
  if (!ref_in) {
    ERROR__(("Failed to read input file %s (reference sequence).\n", reference_filename));
    exit(RETURN_INPUT_ERR);
  }
  // browse the file once
  /* how many sequences are there?  */
  /* Expected format: multifasta (one 'header' line, followed by the sequence) */
  // So count headers...
  while (fgets(line, MAX_LINE_LEN, ref_in)) {
    if (get_input_line_type(line) == LINE_TYPE_SEQ_HEADER) {
      ++sequence_count;
    }
  }

  if (errno) {
    perror("Error when reading the reference (perror) ");
  }

  /* Allocate */
  SAFE_FAILURE__ALLOC(*db, sequence_count, ReferenceDBType);
  
  /* Go back to the beginning and start loading */
  fseek (ref_in, 0, SEEK_SET);
  
  for (i = 0; i < sequence_count; ++i) {
    (*db)[i].masked_regions = NULL;
    
    if (load_reference_sequence(ref_in, &((*db)[i])) <= 0) {
      /* something strange happened, stop loading and return what we have */
      sequence_count = i;
      break;
    }
  }
  fclose(ref_in);
  
  return sequence_count;
}

/**
 * Destroys a reference database
 */
void clear_reference_db(ReferenceDBType* db) {
  while (db->masked_regions != NULL) {
    MaskedRegionType* tmp = db->masked_regions;
    db->masked_regions = db->masked_regions->next;
    free(tmp);
  }
  if (db->name) {
    free(db->name);
    db->name = NULL;
  }
  free(db->sequence);
  db->sequence = NULL;
}

==> storm-sources-0.004.tar.gz: SToRM/reference_data.c <==
#include "reference_data.h"
#include <ctype.h>
#include <unistd.h>
#include <errno.h> 

#include "alignment.h" /* for INDEL_COUNT_LIMIT */

int IGNORE_LOWERCASE = 0;

/**
 * Load reference sequence
 */

inline int load_reference_sequence(FILE* ref_in, ReferenceDBType* db) {
  char line[MAX_LINE_LEN];

  db->name = NULL;

  /* 1) skip (but keep the name of) the fasta header... */
  if (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) == LINE_TYPE_SEQ_HEADER) {
    int name_len = strcspn(line + 1, " \r\n");
    if (name_len > 0) {
      SAFE_FAILURE__ALLOC(db->name, (name_len + 1), char);
      strncpy(db->name, line + 1, name_len);
      if (db->name[0] == '=' || db->name[0] == '*') {
	/* Not allowed as first char in SN : arbitrary replace with '-' */
	db->name[0] = '-';
      }
      db->name[name_len] = '\0';
    } else {
      ERROR__(("Empty fasta header line in the reference sequence.\n"));
      return RETURN_INPUT_ERR;
    }
  } else {
    ERROR__(("Invalid fasta header line in the reference sequence.\n"));
    return RETURN_INPUT_ERR;
  }


  /* 2) keep the start of the nucleotide sequence to come back */
  int i_start = ftell(ref_in);
  int size = 0;

  /* 3) read nucleotide data to measure its size */
  if (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) != LINE_TYPE_SEQ_HEADER) {
    // this is the first line of the sequence
    int line_len = strlen(line);
    
    /* ignore the newline */
    while (line_len > 0 && isspace(line[line_len - 1])) {
      --line_len;
    }

    /* when the first full line of data is empty : error */
    if (line_len <= 0) {
      ERROR__(("Empty dataline in the reference sequence.\n"));
      return RETURN_INPUT_ERR;
    }


    /**/
    
    /* 3.1) full nucleotide data loop to measure the size */
    while (fgets(line, MAX_LINE_LEN, ref_in) && get_input_line_type(line) != LINE_TYPE_SEQ_HEADER) {
      line_len = strlen(line);
      /* ignore the newline */
      while (line_len > 0 && isspace(line[line_len - 1])) {
	--line_len;
      }
      size += line_len;
    }

    /* 3.2) Allocate the estimated size (>= true size) */
    SAFE_FAILURE__ALLOC(db->sequence, COMPRESSED_LEN_N_BYTES(size+2*(INDEL_COUNT_LIMIT)+1,N_BYTES), CODE_TYPE);
    memset(db->sequence,'\0',COMPRESSED_LEN_N_BYTES(size+2*(INDEL_COUNT_LIMIT)+1,N_BYTES));


    /* 3.3) Go back to first base and load the nucleotides sequence */
    fseek(ref_in, i_start, SEEK_SET);

    char c ;
    int  i = 0;
#ifndef NUCLEOTIDES
    CODE_TYPE code_prev = 0;
#endif

    int masked_region_start = -1;
    MaskedRegionType* crt_masked_region = db->masked_regions;

    while ((c = (IGNORE_LOWERCASE ? fgetc(ref_in) : toupper(fgetc(ref_in)))) != EOF) {
      CODE_TYPE code = 0;
      
      if (isspace(c)) {
	continue;
      }

      if (c == SEQ_HEADER_MARKER) {
	/* started to read the next sequence; getting out */
	fseek(ref_in, -1, SEEK_CUR);
	break;
      }

      if (!IS_VALID_BASE_SYMBOL(c)) {
#ifndef NUCLEOTIDES
	code = code_prev; /* masked regions, eg 'NNNNNNNNN' cannot be coded; instead, GNNNNNNNN is transformed in GGGGGGGGG */
#endif
	if (masked_region_start == -1) {
	  /* beginning of a new masked region */
	  masked_region_start = i;
	}
      } else {
	code = BASE_SYMBOL_TO_CODE(c);
	if (masked_region_start != -1) {
	  /* a masked region ends here */
	  if (crt_masked_region == NULL) {
	    /* this is the first one */
	    SAFE_FAILURE__ALLOC(crt_masked_region, 1, MaskedRegionType);
	    db->masked_regions = crt_masked_region;
	  } else {
	    SAFE_FAILURE__ALLOC(crt_masked_region->next, 1, MaskedRegionType);
	    crt_masked_region = crt_masked_region->next;
	  }
	  crt_masked_region->next  = NULL;
	  crt_masked_region->start = masked_region_start;
	  crt_masked_region->end   = i;
	  masked_region_start      = -1;
	}
      }

#ifdef NUCLEOTIDES
      TO_NTH_CODE(db->sequence, i, code);
      i++;
#else
      if (db->first_base == 0xff) {
	code_prev = db->first_base = code;
      } else {
	TO_NTH_CODE(db->sequence, i, COLOR(code_prev, code));
	i++;
	code_prev = code;
      }
#endif      
    }
    db->size = i;
    return db->size;
  } else {
    ERROR__(("Empty fasta block in the reference sequence.\n"));
    return RETURN_INPUT_ERR;
  }
}

/**
 * Load reference sequence
 */
int load_reference_db(const char* reference_filename, ReferenceDBType** db) {
  char line[MAX_LINE_LEN];
  FILE* ref_in = fopen(reference_filename, "r");
  int i=0;
  int sequence_count = 0;
  
  if (!ref_in) {
    ERROR__(("Failed to read input file %s (reference sequence).\n", reference_filename));
    exit(RETURN_INPUT_ERR);
  }
  // browse the file once
  /* how many sequences are there?  */
  /* Expected format: multifasta (one 'header' line, followed by the sequence) */
  // So count headers...
  while (fgets(line, MAX_LINE_LEN, ref_in)) {
    if (get_input_line_type(line) == LINE_TYPE_SEQ_HEADER) {
      ++sequence_count;
    }
  }

  if (errno) {
    perror("Error when reading the reference (perror) ");
  }

  /* Allocate */
  SAFE_FAILURE__ALLOC(*db, sequence_count, ReferenceDBType);
  
  /* Go back to the beginning and start loading */
  fseek (ref_in, 0, SEEK_SET);
  
  for (i = 0; i < sequence_count; ++i) {
    (*db)[i].masked_regions = NULL;
    (*db)[i].name           = NULL;
    (*db)[i].size           = 0;
    (*db)[i].sequence       = NULL;
#ifndef NUCLEOTIDES
    (*db)[i].first_base     = 0xff;
#endif
    if (load_reference_sequence(ref_in, &((*db)[i])) <= 0) {
      /* something strange happened, stop loading and return what we have */
      sequence_count = i;
      break;
    }
  }
  fclose(ref_in);
  
  return sequence_count;
}

/**
 * Destroys a reference database
 */
void clear_reference_db(ReferenceDBType* db) {
  while (db->masked_regions != NULL) {
    MaskedRegionType* tmp = db->masked_regions;
    db->masked_regions = db->masked_regions->next;
    free(tmp);
  }
  if (db->name) {
    free(db->name);
    db->name = NULL;
  }
  free(db->sequence);
  db->sequence = NULL;
}
